@@ -23,17 +23,18 @@ import random as rand  # for random numbers
 
 import matplotlib
 import matplotlib.pyplot as plt  # for plotting
+from matplotlib.ticker import (MultipleLocator, AutoMinorLocator) # control of axes ticks
 import numpy as np  # for ndarrays
 import pandas as pd  # for DataFrames
 from scipy import signal
+from tqdm import tqdm # progress bar
 
 # Hard coded image file output
 image_type = "tif"
 dpi = 600
 
-
-def ndarray2GSLIB(array, data_file, col_name):
-    """Convert 1D or 2D numpy ndarray to a GSLIB Geo-EAS file for use with
+def ndarray2GSLIB_3D(array, data_file, col_name):
+    """Convert 1D or 2D or 3D numpy ndarray to a GSLIB Geo-EAS file for use with
     GSLIB methods.
 
     :param array: input array
@@ -42,14 +43,22 @@ def ndarray2GSLIB(array, data_file, col_name):
     :return: None
     """
 
-    if array.ndim not in [1, 2]:
-        raise ValueError("must use a 2D array")
+    if array.ndim not in [1, 2,3]:
+        raise ValueError("must use a 3D array")
 
     with open(data_file, "w") as f:
         f.write(data_file + "\n")
         f.write("1 \n")
         f.write(col_name + "\n")
 
+        if array.ndim == 3:
+            nz, ny, nx = array.shape
+
+            for iz in range(nz):
+                for iy in range(ny):
+                    for ix in range(nx):
+                        f.write(str(array[nz - 1 - iz,ny - 1 - iy, ix]) + "\n")
+        
         if array.ndim == 2:
             ny, nx = array.shape
 
@@ -98,6 +107,51 @@ def GSLIB2ndarray(data_file, kcol, nx, ny):
                 array[ix] = head.split()[kcol]
     return array, col_name
 
+def GSLIB2ndarray_3D(data_file, kcol,nreal, nx, ny, nz):
+    """Convert GSLIB Geo-EAS file to a 1D or 2D numpy ndarray for use with
+    Python methods
+
+    :param data_file: file name
+    :param kcol: name of column which contains property
+    :param nreal: Number of realizations
+    :param nx: shape along x dimension
+    :param ny: shape along y dimension
+    :param nz: shape along z dimension
+    :return: ndarray, column name
+    """
+    if nz > 1 and ny > 1:
+        array = np.ndarray(shape = (nreal, nz, ny, nx), dtype=float, order="F")
+    elif ny > 1:
+        array = np.ndarray(shape=(nreal, ny, nx), dtype=float, order="F")
+    else:
+        array = np.zeros(nreal, nx)
+
+    with open(data_file) as f:
+        head = [next(f) for _ in range(2)]  # read first two lines
+        line2 = head[1].split()
+        ncol = int(line2[0])  # get the number of columns
+
+        for icol in range(ncol):  # read over the column names
+            head = next(f)
+            if icol == kcol:
+                col_name = head.split()[0]
+        for ineal in range(nreal):
+            if nz > 1 and ny > 1:
+                for iz in range(nz):
+                    for iy in range(ny):
+                        for ix in range(nx):
+                            head = next(f)
+                            array[ineal][nz - 1 - iz][ny - 1 - iy][ix] = head.split()[kcol]
+            elif ny > 1:
+                for iy in range(ny):
+                    for ix in range(0, nx):
+                        head = next(f)
+                        array[ineal][ny - 1 - iy][ix] = head.split()[kcol]
+            else:
+                for ix in range(nx):
+                    head = next(f)
+                    array[ineal][ix] = head.split()[kcol]
+    return array, col_name
 
 def Dataframe2GSLIB(data_file, df):
     """Convert pandas DataFrame to a GSLIB Geo-EAS file for use with GSLIB
@@ -814,6 +868,70 @@ def locpix_log_st(
     cbar.set_label(vlabel, rotation=270, labelpad=20)
     return cs
 
+def conditioning_check(model,nx,xmn,xsiz,ny,ymn,ysiz,nreal,df,xcol,ycol,vcol,vname,vmin,vmax):
+    
+    iix = np.zeros(len(df),dtype = int); iiy = np.zeros(len(df),dtype = int)
+    for idata in range(0,len(df)):                                # find the model ix, iy cells at data x, y
+        iix[idata] = max(0,min(int((df.loc[idata,xcol] - xmn)/xsiz),nx-1))
+        iiy[idata] = ny - max(0,min(int((df.loc[idata,ycol] - ymn)/ysiz),ny-1))-1
+        
+    real_por = np.zeros(len(df))
+    all_data = np.zeros(len(df)*nreal)
+    all_real = np.zeros(len(df)*nreal)
+    
+    i = 0
+    for ireal in tqdm(range(0,nreal)):                            # make an array of all realizations paired to data values
+        for idata in range(0,len(df)): 
+            real_por[idata] = model[ireal,iiy[idata],iix[idata]]  
+            all_data[i] = df.loc[idata,vcol]
+            all_real[i] = real_por[idata]
+            i = i + 1
+        df['Real' + str(ireal+1)] = real_por 
+    
+    df_vstack = pd.DataFrame(np.vstack((all_data, all_real)).T,columns = ['Data','Real']) # convert to a DataFrame
+        
+    plt.scatter(all_data,all_real,alpha=0.2,color='black',s=5)
+    plt.plot([vmin,vmax],[vmin,vmax],color='black',linestyle='--')
+    plt.ylabel('Simulated Realizations - ' + vname); plt.xlabel('Data Values - ' + vname); plt.title('Data Reproduction Check - ' + vname)
+    plt.xlim([vmin,vmax]); plt.ylim([vmin,vmax])
+    return df_vstack
+
+def distribution_check(df,vcol,model,nx,ny,nreal,dreal,vname,vmin,vmax):
+    boxprops = dict(linewidth=1.0, color='black')
+    whiskerprops = dict(linestyle='-',linewidth=1.0, color='black')
+    medianprops = dict(linestyle='--',linewidth=1.0, color='black')
+    meanprops={"marker":'_', 'markersize': 15,"markerfacecolor":"red", "markeredgecolor":'red'}
+    flierprops={'marker': 'o', 'markersize': 4, 'markerfacecolor': 'grey','markeredgecolor':'black'}
+    
+    refdist = df[vcol].values
+    
+    fig, axs = plt.subplots(1,2,figsize=(16,5), gridspec_kw={'width_ratios': [1, 10]}, sharey = True)
+    
+    axs[0].boxplot(refdist,labels = [''],whis=True,showmeans=True,widths=0.3,notch=True,boxprops=boxprops, whiskerprops=whiskerprops,meanprops=meanprops,flierprops=flierprops,medianprops=medianprops)
+    axs[0].set_ylim([0,28]); axs[0].set_ylabel(vname); axs[0].set_xlabel('Input Data'); axs[0].set_title('')
+    axs[0].grid(True,alpha=0.3,axis='y'); axs[0].set_xlim([1-0.4,1+0.4])
+    axs[0].plot([1-0.4,1+0.4],[np.percentile(refdist,75),np.percentile(refdist,75)],color='black',alpha=0.2,linestyle='--',linewidth=2)
+    axs[0].plot([1-0.4,1+0.4],[np.percentile(refdist,50),np.percentile(refdist,50)],color='black',alpha=0.2,linestyle='--',linewidth=2)
+    axs[0].plot([1-0.4,1+0.4],[np.percentile(refdist,25),np.percentile(refdist,25)],color='black',alpha=0.2,linestyle='--',linewidth=2)
+    axs[0].yaxis.set_minor_locator(plt.MultipleLocator(20))   
+    axs[0].plot([1-0.4,1+0.4],[np.average(refdist),np.average(refdist)],color='red',alpha=0.2,linestyle='--',linewidth=2)
+    
+    axs[0].yaxis.grid(True, which='major',linewidth = 2.0); axs[0].yaxis.grid(True, which='minor',linewidth = 0.1) # add y grids
+    axs[0].tick_params(which='major',length=7); axs[0].tick_params(which='minor', length=4)
+    axs[0].yaxis.set_minor_locator(AutoMinorLocator()) # turn on minor ticks
+    tarray = model.reshape(nreal,ny*nx).T
+    axs[1].boxplot(tarray[:,:dreal],whis=True,showmeans=True,widths=0.2,notch=True,boxprops=boxprops, whiskerprops=whiskerprops,meanprops=meanprops,flierprops=flierprops,medianprops=medianprops)
+    axs[1].set_ylim([vmin,vmax]); axs[1].set_ylabel(''); axs[1].set_xlabel('Realizations'); axs[1].set_title('Check Global ' + vname + ' Distribution - Summary Statistics')
+    axs[1].grid(True,alpha=0.3,axis='y'); axs[1].set_xlim([1-0.2,dreal+0.2])
+    axs[1].plot([1-0.2,dreal+0.2],[np.percentile(refdist,75),np.percentile(refdist,75)],color='black',alpha=0.2,linestyle='--',linewidth=2)
+    axs[1].plot([1-0.2,dreal+0.2],[np.percentile(refdist,50),np.percentile(refdist,50)],color='black',alpha=0.2,linestyle='--',linewidth=2)
+    axs[1].plot([1-0.2,dreal+0.2],[np.percentile(refdist,25),np.percentile(refdist,25)],color='black',alpha=0.2,linestyle='--',linewidth=2)
+    axs[1].plot([1-0.2,dreal+0.2],[np.average(refdist),np.average(refdist)],color='red',alpha=0.2,linestyle='--',linewidth=2)
+    
+    axs[1].yaxis.grid(True, which='major',linewidth = 2.0); axs[1].yaxis.grid(True, which='minor',linewidth = 0.1) # add y grids
+    axs[1].tick_params(which='major',length=7); axs[1].tick_params(which='minor', length=4)
+    axs[1].yaxis.set_minor_locator(AutoMinorLocator()) # turn on minor ticks
+    return axs
 
 def affine(array, tmean, tstdev):
     """Affine distribution correction reimplemented in Python with numpy
@@ -1897,25 +2015,26 @@ def sgsim_3D(nreal, df, xcol, ycol, zcol, vcol, nx, ny, nz, hsiz, vsiz, seed, va
 
     nug = var["nug"]
     nst = var["nst"]
-    it1 = var["it1"]
-    cc1 = var["cc1"]
-    azi1 = var["azi1"]
-    dip1 = var["dip1"] 
-    hmax1 = var["hmax1"]
-    hmed1 = var["hmed1"]
-    hmin1 = var["hmin1"]
-    it2 = var["it2"]
-    cc2 = var["cc2"]
-    azi2 = var["azi2"]
-    dip2 = var["dip2"] 
-    hmax2 = var["hmax2"]
-    hmed2 = var["hmed2"]
-    hmin2 = var["hmin2"]
+    it1 = var["it"][0]
+    cc1 = var["cc"][0]
+    azi1 = var["azi"][0]
+    dip1 = var["dip"][0] 
+    hmax1 = var["hmaj"][0]
+    hmin1 = var["hmin"][0]
+    hvert1 = var["hvert"][0]
+    it2 = var["it"][1]
+    cc2 = var["cc"][1]
+    azi2 = var["azi"][1]
+    dip2 = var["dip"][1] 
+    hmax2 = var["hmaj"][1]
+    hmin2 = var["hmin"][1]
+    hvert2 = var["hvert"][1]
     max_range = max(hmax1, hmax2)
-    max_range_v = max(hmin1, hmin2) * 10
+    max_range_v = max(hvert1, hvert2)
     hmn = hsiz * 0.5
     zmn = vsiz * 0.5
-    hctab = int(max_range / hsiz) * 2 + 1
+    hctab_h = int(max_range / hsiz) * 2 + 1
+    hctab_v = int(max_range_v / vsiz) * 2 + 1
 
     with open("sgsim.par", "w") as f:
         f.write("              Parameters for SGSIM                                         \n")
@@ -1948,7 +2067,7 @@ def sgsim_3D(nreal, df, xcol, ycol, zcol, vcol, nx, ny, nz, hsiz, vsiz, seed, va
         f.write("0                             -maximum data per octant (0=not used)        \n")
         f.write(str(max_range) + " " + str(max_range) +" "+ str(max_range_v) + " -maximum search  (hmax,hmin,vert) \n")
         f.write(str(azi1) + "   0.0   0.0       -angles for search ellipsoid                 \n")
-        f.write(str(hctab) + " " + str(hctab) + " 1 -size of covariance lookup table        \n")
+        f.write(str(hctab_h) + " " + str(hctab_h) + " " + str(hctab_v) + " -size of covariance lookup table        \n")
         f.write("1     0.60   1.0              - ktype: 0=SK,1=OK,2=LVM,3=EXDR,4=COLC        \n")
         f.write("none.dat                      -  file with LVM, EXDR, or COLC variable     \n")
         f.write("4                             -  column for secondary variable             \n")
@@ -1961,50 +2080,3 @@ def sgsim_3D(nreal, df, xcol, ycol, zcol, vcol, nx, ny, nz, hsiz, vsiz, seed, va
     os.system("sgsim.exe sgsim.par")
     sim_array = GSLIB2ndarray_3D(output_file, 0, nreal, nx, ny, nz)
     return sim_array[0]
-
-
-def GSLIB2ndarray_3D(data_file, kcol,nreal, nx, ny, nz):
-    """Convert GSLIB Geo-EAS file to a 1D or 2D numpy ndarray for use with
-    Python methods
-
-    :param data_file: file name
-    :param kcol: name of column which contains property
-    :param nreal: Number of realizations
-    :param nx: shape along x dimension
-    :param ny: shape along y dimension
-    :param nz: shape along z dimension
-    :return: ndarray, column name
-    """
-    if nz > 1 and ny > 1:
-        array = np.ndarray(shape = (nreal, ny, nx, nz), dtype=float, order="F")
-    elif ny > 1:
-        array = np.ndarray(shape=(nreal, ny, nx), dtype=float, order="F")
-    else:
-        array = np.zeros(nreal, nx)
-
-    with open(data_file) as f:
-        head = [next(f) for _ in range(2)]  # read first two lines
-        line2 = head[1].split()
-        ncol = int(line2[0])  # get the number of columns
-
-        for icol in range(ncol):  # read over the column names
-            head = next(f)
-            if icol == kcol:
-                col_name = head.split()[0]
-        for ineal in range(nreal):		
-            if nz > 1 and ny > 1:
-                for iz in range(nz):
-                    for iy in range(ny):
-                        for ix in range(nx):
-                            head = next(f)
-                            array[ineal][ny - 1 - iy][ix][iz] = head.split()[kcol]    					
-            elif ny > 1:
-                for iy in range(ny):
-                    for ix in range(0, nx):
-                        head = next(f)
-                        array[ineal][ny - 1 - iy][ix] = head.split()[kcol]
-            else:
-                for ix in range(nx):
-                    head = next(f)
-                    array[ineal][ix] = head.split()[kcol]
-    return array, col_name
\ No newline at end of file
